/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package cricketdaemon.clientlib.data;

import cricketdaemon.clientlib.*;
import java.util.*;

/**
 * Contains information update dispatched by Broker.  Application uses
 * the public methods here to extract the available information.  A mask
 * is provided to indicate which Broker.TYPE of information is
 * available.
 *
 * Beware when multiple threads have registered callbacks with the
 * Broker: the same <code>CricketData</code> gets passed to every
 * thread. As a result, any modifications made to data may become
 * visible to another thread.  In such case, each thread is responsible
 * for deep cloning the relevant data structure.  Be advised that deep
 * cloning is not implemented for any data strucuture contained within
 * <code>CricketData</code>.  
 *
 * @author Allen Miu
 */
public class CricketData {
    private int seqNum;

    /* BeaconRecord of the current space */
    private BeaconRecord currentSpaceObject;

    /* array list of BeaconRecords within CricketDaemon's sliding window */
    private ArrayList beaconsHeard;

    /* array list of BeaconRecords containing distance estimates to each
       beacon.  The values are generated as part of the least squares
       computation used for solving the listener's position */
    private ArrayList calcBeaconDist;

    /* current x,y,z position */
    private Position devicePosition;

    /* id of the algorithm used to compute the devicePosition */
    private String posAlgor; 

    /* anchor beacon for orientation. Orientation is expressed as the relative angle 
       from the line drawn between the listener and the anchorBeacon
    */
    private BeaconRecord anchorBeacon;

    /* angle in radians */
    private double orient;

    /* a bit is set for the valid data type included in this CricketData object */
    private BitSet mask;

    // time that ServerBroker receives data from CricketDaemon
    private long timeStamp;

    // Cricket hw clock stamp of the latest received beacon in CricketData
    private long hwLastUpdate = Long.MIN_VALUE;

   
    public CricketData(long time) 
    {
	mask = new BitSet();
	timeStamp = time;
	orient = 0;
    }


    public String getCurrentSpace()
    {
	BeaconRecord nearest = getCurrentSpaceObject();
	if(nearest != null)
	    return nearest.space;
	return null;
    }
    public ArrayList getBeaconsHeard() { return beaconsHeard; }
    public ArrayList getCalcBeaconDist() { return calcBeaconDist; }
    public BeaconRecord getCurrentSpaceObject() { return currentSpaceObject; }
    public Position getDevicePosition() { return devicePosition; }
    public String getPosAlgor() { return posAlgor; }
    public long getHWTimeStamp() { return hwLastUpdate; }
    public BitSet getMask() { return mask; }
    public int getSeqNum() { return seqNum; }
    public long getTimeStamp() { return timeStamp; }


    // cricketdaemon does not yet process orientation information
    public BeaconRecord getAnchorBeacon() { return anchorBeacon; }
    public double getOrient() { return orient; }

    public String toString()
    {
	String result = "CricketData: "+seqNum+" "+mask+"\n"+
	    "  currentSpaceObject: "+currentSpaceObject+"\n";
	if(beaconsHeard != null) {
	    result += "  beaconsHeard: [ ";
	    for(int i = 0; i < beaconsHeard.size(); i++) 
		result += beaconsHeard.get(i)+" ::: ";
	    result += " ]\n";
	}
	result += "  devicePosition: "+devicePosition;
	if(calcBeaconDist != null) {
	    result += "  calcBeaconDist: [ ";
	    for(int i = 0; i < calcBeaconDist.size(); i++) 
		result += calcBeaconDist.get(i)+" ::: ";
	    result += " ]\n";
	}
	result += "  anchorBeacon:   "+anchorBeacon;
	result += "  orientation:    "+orient;
	return result;
    }
    
    // all the methods below are used by the Parser and not normally 
    // used by clients

    // used to combine two CricketData into one CricketData. 
    public void merge(CricketData other)
    {
	// FIXME: what should be done with seqNum?
	if(currentSpaceObject == null)
	    currentSpaceObject = other.currentSpaceObject;
	if(devicePosition == null)
	    devicePosition = other.devicePosition;
	if(anchorBeacon == null)
	    anchorBeacon = other.anchorBeacon;
	if(orient == 0)
	    orient = other.orient;
	mask.or(other.mask);

	if(beaconsHeard == null)
	    beaconsHeard = other.beaconsHeard;
	else if(other.beaconsHeard != null) {
	    beaconsHeard.addAll(other.beaconsHeard);
	}

	posAlgor = other.getPosAlgor();

	if(calcBeaconDist == null)
	    calcBeaconDist = other.calcBeaconDist;
	else if(other.calcBeaconDist != null) {
	    calcBeaconDist.addAll(other.calcBeaconDist);
	}

	if(other.hwLastUpdate > this.hwLastUpdate)
	    this.hwLastUpdate = other.hwLastUpdate;
    }


    /* FIXME: seqNum not really used now */
    public void setSeqNum(int c) 
    { 
	mask.set(Broker.SEQNUM);
	seqNum = c; 
    }

    public void setCurrentSpaceObject(BeaconRecord c) 
    { 
	if(c != null) {
	    mask.set(Broker.SPACE);
	    currentSpaceObject = c; 
	}
    }
    public void setBeaconsHeard(ArrayList c) 
    { 
	if(c != null) {
	    mask.set(Broker.BEACONSHEARDLIST);
	    beaconsHeard = c; 
	    //setCurrentSpaceObject((BeaconRecord) beaconsHeard.get(0));

	    Iterator it = beaconsHeard.iterator();
	    while(it.hasNext()) {
		BeaconRecord br = (BeaconRecord) it.next();
		if(br.lastUpdate.longValue() > hwLastUpdate)
		    hwLastUpdate = br.lastUpdate.longValue();
	    }
	}
    }

    public void setCalcBeaconDist(ArrayList a) 
    {
	if(a != null) {
	    mask.set(Broker.CALCBEACONDIST);
	    calcBeaconDist = a;
	}
    }

    public void setPosAlgor(String s) {	posAlgor = s; }

    public void setDevicePosition(Position c) 
    { 
	if(c != null) {
	    if(c.invalid())
		mask.set(Broker.POSERR);
	    else
		mask.set(Broker.DEVICECOORD);
	    devicePosition = c; 
	}
    }

    public void setOrient(double c) 
    { 
	mask.set(Broker.ORIENTATION);
	orient = c; 
    }

    public void setAnchorBeacon(BeaconRecord c) 
    { 
	if(c != null) {
	    mask.set(Broker.ANCHORBEACON);
	    anchorBeacon = c; 
	}
    }

    /* Currently, if the beaconsheard list is available, beaconstat is
       automatically available */
    public void setBeaconStat() {
	mask.set(Broker.BEACONSTAT); 
    }

}
