/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package cricketdaemon.position;

import cricketdaemon.reader.Beacon;
import cricketdaemon.reader.CartesianCoord;
import cricketdaemon.space.DistanceEstimate;
import cricketdaemon.Util;
import Jama.*;
import java.util.*;

public class EKFSolver
{
    class AccumBuffer
    {
	private int accum;
	private ArrayList l;
	
	AccumBuffer(int v, int len)
	{
	    l = new ArrayList();
	    for(int i = 0; i < len; i++) {
		l.add(new Integer(v));
		accum += v;
	    }
	}

	// operations not synchronized so it's dangerous to have multiple read/write 
	// threads to access the same instance of this object!
	public int insert(int v) {
	    // assumes invariant is l.size() >= 1
	    Integer r = (Integer) l.remove(0);
	    l.add(new Integer(v));
	    
	    accum -= r.intValue();
	    accum += v;
	    
	    return r.intValue();
	}

	public int sum()
	{
	    return accum;
	}

	public double ratio()
	{
	    return (double) sum()/size();
	}

	public int size()
	{
	    return l.size();
	}
    }

    class EKFSample
    {
	Beacon beacon;
	String space;
	int id;
	CartesianCoord coord;
	double dist;
	long time;

	EKFSample(Beacon b, String s, int i, CartesianCoord c, double d, long t)
	{
	    beacon = b;
	    space = s; id = i; coord = c; dist = d; time = t;
	}

	public boolean equals(Object o)
	{
	    if(!(o instanceof EKFSample))
		return false;
	    EKFSample ekfso = (EKFSample) o;
	    return space.equals(ekfso.space) && id == ekfso.id;
	}

	public String toString()
	{
	    return space+"%"+id+" "+coord+" "+dist+" "+time;
	}
    }

    private static final int OK            = 0;
    private static final int REJECT_SAMPLE = 1;
    private static final int BAD_STATE     = 2;

    private Matrix X;   // 1xDIM x, y, z, vx, vy, vz
    private Matrix D;   // 1xDIM 
    private Matrix U;   // DIMxDIM
    private Matrix G;   // DIMxDIM

    private Matrix xEst; // 1xDIM
    private Matrix dEst;
    private Matrix uEst;

    private double fCov;

    private static final double R = 145.2218;
    private static final double rejectThresh = 1000;
    private static final double suspectThresh = Double.POSITIVE_INFINITY;
    private static final double rejectRatio = 0.4;
    private double suspectRatio = 0.2;
    private double qScale = Math.pow(10, -4.4);
    private double UVAL = 6.5567;
    private static final double DELAY_THRESH = 1000;
    private static final double MULTI_NUM = 2;

    private static final int BEACONBUF_SIZE = 4;

    private ArrayList beaconBuf;

    private AccumBuffer suspectBuf;
    private AccumBuffer rejectBuf;

    private long lastUpdate = 0;
    private long runCounter = 0;  // potential wrap around problem

    //sqrt3 = 1.732050808;
    public final double SQRT3 = Math.sqrt(3);


    private boolean threeStatesEKF = false;
    private int DIM = 6; // number of EKF states

    private EKFSample lastEKFSample = null;

    ////// Some Utililty methods //////

    // debug
    public static void dumpMat(String label, Matrix x) 
    {
	System.out.println(label);
	x.print(x.getColumnDimension(), 5);
    }

    //// copies diagonal of mxm Matrix S to a 1xm or mx1 matrix T.
    public static void copyDiag(Matrix t, Matrix s)
    {
	int tm = t.getRowDimension();
	int tn = t.getColumnDimension();

	int sm = s.getRowDimension();
	int sn = s.getColumnDimension();

	if(sm != sn)
	    return;

	if(tm == 1) {
	    if(tn != sm)
		return;

	    for(int i = 0; i < tn; i++)
		t.set(0, i, s.get(i, i));
	}
	else if(tn == 1) {
	    if(tm != sm)
		return;
	    for(int i = 0; i < tm; i++)
		t.set(i, 0, s.get(i, i));
	}
    }

    public static Matrix diag(Matrix d)
    {
	int dm = d.getRowDimension();
	int dn = d.getColumnDimension();

	Matrix result = null;

	if(dm == 1 || dn == 1) {
	    // expand
	    if(dm > dn) {
		result = new Matrix(dm, dm);
		for(int i = 0; i < dm; i++)
		    result.set(i, i, d.get(i, 0));
	    }
	    else {
		result = new Matrix(dn, dn);
		for(int i = 0; i < dn; i++)
		    result.set(i, i, d.get(0, i));
	    }

	}
	else {
	    // contract
	    int dim = Util.min(dm, dn);
	    result = new Matrix(1, dim);
	    for(int i = 0; i < dim; i++)
		result.set(0, i, d.get(i, i));
	}

	return result;
    }

    // returns new dxd matrix filled with value v in the diagonal and 0 elsewhere
    private Matrix newEye(int d, double v)
    {
	return Matrix.identity(d,d).timesEquals(v);
    }

    
    private Matrix newU()
    {
	//return newEye(6, Math.sqrt(3*qScale));
	return newEye(DIM, UVAL);
    }

    // returns new 1xd Matrix filled with with value v
    private Matrix newOneByD(int d, double v)
    {
	return diag(Matrix.identity(d, d)).timesEquals(v);
    }

    public EKFSolver(boolean threeStates)
    {
	this.threeStatesEKF = threeStates;

	if(threeStatesEKF) {
	    suspectRatio = 0.4;
	    UVAL = 3.5567;
	    qScale = Math.pow(10, -4);
	    DIM = 3;
	}

	//X = new Matrix(1, DIM);
	X = null;
	xEst = new Matrix(1, DIM);
	G = new Matrix(DIM, DIM);
	D = newOneByD(DIM, 1);
	U = newU();

	beaconBuf  = new ArrayList();
	suspectBuf = new AccumBuffer(0, 20);
	rejectBuf  = new AccumBuffer(0, 20);


    }

    private void updateG(long dT)
    {
	double sqrtdt = Math.sqrt(dT);
	double G1 = (SQRT3/3)*sqrtdt*dT;
	Util.setDiagMatrix(G, 0, 0, 3, G1);

	if(!threeStatesEKF) {
	    double G2 = SQRT3/2*sqrtdt;
	    double G3 = (1/2)*sqrtdt;
	    Util.setDiagMatrix(G, 0, 3, 3, G2);
	    Util.setDiagMatrix(G, 3, 0, 3, 0);
	    Util.setDiagMatrix(G, 3, 3, 3, G3);
	}
    }

    private void statePrediction(long dT)
    {
	if(X == null)
	    return;

	if(threeStatesEKF) {	
	    xEst.set(0,0, X.get(0,0));
	    xEst.set(0,1, X.get(0,1));
	    xEst.set(0,2, X.get(0,2));
	}
	else {
	    xEst.set(0,0, X.get(0,0)+X.get(0,3)*dT);
	    xEst.set(0,1, X.get(0,1)+X.get(0,4)*dT);
	    xEst.set(0,2, X.get(0,2)+X.get(0,5)*dT);
	    xEst.set(0,3, X.get(0,3));
	    xEst.set(0,4, X.get(0,4));
	    xEst.set(0,5, X.get(0,5));
	}
    }

    private boolean reject(double chisqr)
    {
	return Double.isNaN(chisqr) || chisqr > rejectThresh;
    }

    private int covarPredictionAndCorrection(long dT, EKFSample s)
    {
	if(dT > DELAY_THRESH || X == null) // need to bootstrap
	    return BAD_STATE;

	Matrix myG  = G.copy();
	Matrix phiU = U.copy(); // rows of [PhiU,G] are to be orthononalized

	uEst = Matrix.identity(DIM, DIM);// initialize lower triangular part of U
	dEst = new Matrix(1,DIM);

	if(!threeStatesEKF) {
	    phiU.set(0,3, phiU.get(0,3) + dT*phiU.get(3,3));
	    phiU.set(0,4, phiU.get(0,4) + dT*phiU.get(3,4));
	    phiU.set(0,5, phiU.get(0,5) + dT*phiU.get(3,5));
	    phiU.set(1,4, phiU.get(1,4) + dT*phiU.get(4,4));
	    phiU.set(1,5, phiU.get(1,5) + dT*phiU.get(4,5));
	    phiU.set(2,5, phiU.get(2,5) + dT*phiU.get(5,5));
	}
	
	for(int ivar = DIM-1; ivar >=0; ivar -= 1) {
	    double sigma = 0;
	    for(int  k = ivar; k <= DIM-1; k += 1) {
		sigma += CartesianCoord.sqr(phiU.get(ivar,k))*D.get(0,k);
		sigma += CartesianCoord.sqr(myG.get(ivar,k))*qScale; //?
	    }

	    /*
	      if(sigma <= 0.0000001) {
	      System.out.println("ERROR!");
	      dumpMat("phiU", phiU);
	      dumpMat("myG", myG);
	      dumpMat("D", D);
	      dumpMat("dEst", dEst);
	      System.exit(1);
	      }
	    */

	    // if sigma is zero, we have a problem!!?? b in bierman gets NaN
	    dEst.set(0, ivar, sigma);
           
	    for(int j = 0; j < ivar; j += 1) {
		sigma = 0;
		for(int k=ivar; k <= DIM-1; k += 1) {
		    sigma = sigma + phiU.get(ivar,k)*D.get(0,k)*phiU.get(j,k);
		    sigma = sigma + myG.get(ivar,k)*qScale*myG.get(j,k); //?
		}
		uEst.set(j, ivar, sigma/dEst.get(0, ivar));
		
		for(int k=ivar; k <= DIM-1; k += 1) {
		    // This does NOT change zero / non-zero structure of PhiU
		    phiU.set(j,k, (phiU.get(j,k) - U.get(j,ivar)*phiU.get(ivar,k))); 
		    
		    // U(j,i) term is always non-zero by def of loop bounds
		    myG.set(j,k, (myG.get(j,k) - U.get(j,ivar)*myG.get(ivar,k))); 
		}
	    }
	}

	// begion correction step
	Matrix PVH = new Matrix(1, DIM);
	CartesianCoord c     = s.coord;
	double htheo = c.dist(xEst.get(0,0), xEst.get(0,1), xEst.get(0,2));
	
	PVH.set(0,0, (xEst.get(0, 0)- c.x)/htheo);
	PVH.set(0,1, (xEst.get(0, 1)- c.y)/htheo);
	PVH.set(0,2, (xEst.get(0, 2)- c.z)/htheo);

	double chisqr = CartesianCoord.sqr(s.dist - htheo);
	boolean ignoreOutlierRejection = false;
	int badest = 0;
	fCov = 0; // remains 0 if EKF falls into bad state

	// Record chisqr value, must do this before next loop
	if (reject(chisqr)) {
	    rejectBuf.insert(1);
	    suspectBuf.insert(0); // We don't want to count an outlier as suspicious
	}
	else {
	    rejectBuf.insert(0);
	    if (chisqr > suspectThresh)
		suspectBuf.insert(1);
	    else
		suspectBuf.insert(0);
	}
	ignoreOutlierRejection = (rejectBuf.ratio() >= rejectRatio);

	/*
	  System.out.println("measud "+s.dist);
	  System.out.println("htheo  "+htheo);
	  System.out.println("chisqr "+chisqr);
	  System.out.println("iOutRe "+ignoreOutlierRejection);
	  System.out.println("rejBuf "+rejectBuf.sum());
	  System.out.println("susBuf "+suspectBuf.sum());
	*/

	if (reject(chisqr) && (runCounter > 20)) { //throw away sample
	    if (!ignoreOutlierRejection) {
		Util.log("Throwing out sample, ignoring bad state indicator");
		xEst = X.copy();
		//dEst = D.copy();
		//uEst = U.copy();
		return REJECT_SAMPLE;
	    }
	    else {
		badest = 1;
	    }
	}
	else if (suspectBuf.ratio() >= suspectRatio) {
	    badest = 2;
	}
    
	if (badest > 0) 
	    return BAD_STATE;

	// ----END OUTLIER REJECTOR / BAD STATE ESTIMATOR----

	/*
	  System.out.println("R="+R);
	  dumpMat("PVH=", PVH);
	  dumpMat("xEst=", xEst);
	  dumpMat("uEst=", uEst);
	  dumpMat("dEst=", dEst);
	*/
	// ----BEGIN CORRECTOR----
	// dEst should contain the diagonal of the dEst from Bierman process 
	// method only modifies xEst and uEst!!
	double cEst = biermanProcess( s.dist, htheo, R, PVH, xEst, uEst, dEst); // matlab

	/*
	  System.out.println("after bierman");
	  dumpMat("xEst=", xEst);
	  dumpMat("uEst=", uEst);
	*/
	
	X = xEst.copy();
	U = uEst.copy();
	D = dEst.copy();

	// if we're in bad state then f_cov will remain zero
	// fCov used for weighing multi-modal schemes
        fCov = Math.exp(-0.5*chisqr/cEst)/(Math.pow(2*Math.PI*cEst, MULTI_NUM/2));
	
	// for logging purposes only
	//chisqr = chisqr(1); %for now only return one of the chisqr values

	return OK;
    }

    //  Matlab implementation of the
    //  Bierman ``square root filtering without square roots''
    
    //  From the diskette included with
    //  M. S. Grewal, L. R. Weill and A. P. Andrews
    //  Global Positioning Systems, Inertial Navigation and Integration
    //  John Wiley & Sons, 2000.
    
    //  INPUTS:
    //   z    measurement (SCALAR)
    //   R    variance of measurement error
    //   H    measurement sensitivity (row) vector
    //   X  a priori estimate of state vector
    //   U  unit upper triangular factor of covariance matrix of a priori state uncertainty
    //   D  diagonal factor of covariance matrix of a priori state uncertainty
    //  OUTPUTS:
    //   X    a posteriori estimate of state vector
    //   U    upper triangular UD factor of a posteriori state uncertainty covariance
    //   D    diagonal UD factor of a posteriori state uncertainty covariance
    //   cEst  this is "C
    private static double biermanProcess(double z, double zin, double R, Matrix H, Matrix X, Matrix U, Matrix D)
    {
	Matrix dIn = diag(D); 
	Matrix a  = U.transpose().times(H.transpose());    // a is not modified, but
	Matrix b  = dIn.times(a);                          // b is modified to become unscaled Kalman gain.
	double dz = z - zin;
	double alpha = R;
	double gamma = 1/alpha;

	/*
	  System.out.println("dz="+dz);
	  System.out.println("alpha="+alpha);
	  System.out.println("gamma="+gamma);
	  dumpMat("dIn", dIn);
	  dumpMat("a", a);
	  dumpMat("b", b);
	  dumpMat("H", H);
	  dumpMat("U", U);
	*/

	for (int j = 0; j < X.getColumnDimension(); j++) {
	    double beta = alpha;
	    alpha  = alpha + a.get(j, 0)*b.get(j, 0);
	    double lambda = -a.get(j,0)*gamma;
	    gamma  = 1/alpha;
	    dIn.set(j,j,  beta*gamma*dIn.get(j,j));
	    
	    for(int i = 0; i < j-1; i++) {
		beta   = U.get(i,j);
		U.set(i,j,  beta + b.get(i,0)*lambda);
		b.set(i,0,  b.get(i,0) + b.get(j,0)*beta);
	    }
	}
	double dzs = gamma*dz;  // apply scaling to innovations

	X.plusEquals(b.transpose().timesEquals(dzs)); // multiply by unscaled Kalman gain
	copyDiag(D, dIn);
	
	/*
	  System.out.println("dzs="+dzs);
	  dumpMat("b", b);
	  dumpMat("X", X);
	  dumpMat("D", D);
	*/

        return alpha;           // this is "C
    }

    // exposing this method is an ugly hack but this method is now
    // used to generate a 
    // PositionData object with metadata from this solver 
    // and the given x, y, z values
    public PositionData curPos(double x, double y, double z)
    {
	double[] bx = new double[beaconBuf.size()];
	double[] by = new double[beaconBuf.size()];
	double[] bz = new double[beaconBuf.size()];
	double[] d_bar = new double[beaconBuf.size()];
	String[] space = new String[beaconBuf.size()];
	int[] id = new int[beaconBuf.size()];
	CartesianCoord c = new CartesianCoord(x, y, z);
	int i = 0;

	//dumpMat("curPos() X", X);
	
	Iterator it = beaconBuf.iterator();
	while(it.hasNext()) {
	    EKFSample s = (EKFSample) it.next();
	    bx[i] = s.coord.x;
	    by[i] = s.coord.y;
	    bz[i] = s.coord.z;
	    space[i] = s.space;
	    id[i] = s.id;
	    d_bar[i] = c.dist(bx[i], by[i], bz[i]);
	    
	    //System.out.println("EKFSample "+i+": "+s);
	    i++;
	}

	if(lastEKFSample == null)
	    return null;
	else {
	    return new PositionData( lastEKFSample.beacon,
				     c.x, c.y, c.z,
				     bx, by, bz,
				     space, id, d_bar, PositionModule.K_NOMINAL );
	}
    }

    private void insertSample(EKFSample s)
    {
	int i = beaconBuf.indexOf(s);
	if(i >= 0)
	    beaconBuf.remove(i);
	beaconBuf.add(s);
	while(beaconBuf.size() > BEACONBUF_SIZE) 
	    beaconBuf.remove(0);
	lastEKFSample = s;
    }

    private int resetted = 0;  // used for debugging

    public PositionData process(Beacon b, String space, int id, CartesianCoord c, double dist)
    {
	long time = b.hwTimeStamp.longValue();
	long dT = time - lastUpdate;
	lastUpdate = time;
	//System.out.println("dT ="+dT);

	if(dT < 0) // 32-bit clock wrapped around
	    dT += Util.MAX_HWCLOCKVAL;

	EKFSample s = new EKFSample(b, space, id, c, dist, time);
	insertSample(s);

	updateG(dT);
	statePrediction(dT);
	int result = covarPredictionAndCorrection(dT, s);

	//System.out.println("EKF state="+result);

	//if(resetted >= 1 && runCounter >= 2)
	//System.exit(1);

	if (result == OK) {
	    runCounter++;
	    return curPos(X.get(0, 0), X.get(0, 1), X.get(0, 2));
	}

	if (result == REJECT_SAMPLE) {
	    return curPos(X.get(0, 0), X.get(0, 1), X.get(0, 2));
	}

	//case BAD_STATE: or any other state, fall thru and return null
	return null;
    }
    
    // fCov gets updated after calling process() 
    public double getfCov() { return fCov; }

    public void reset(PositionData pos)
    {
	CartesianCoord p = pos.getCoord();
	if(p == null) // error pos, ignore
	    return;

	System.out.println("Bad state detected in non-hybrid mode\n"+
			   "  using LLS under (false) simultaneity assumption");
	xEst.set(0, 0, p.x);
	xEst.set(0, 1, p.y); 
	xEst.set(0, 2, p.z); 
	
	if(!threeStatesEKF) {
	    xEst.set(0, 3, 0);
	    xEst.set(0, 4, 0);
	    xEst.set(0, 5, 0);
	}

	X = xEst.copy();
	D = newOneByD(DIM, 1);
	U = newU();

	//dumpMat("reset() xEst", xEst);

	beaconBuf.clear();
	DistanceEstimate[] d = pos.getDistances();
	for(int i = 0; i < d.length; i++) {
	    insertSample(new EKFSample(pos.beacon, d[i].space, d[i].beaconID, d[i].coord, d[i].distance, lastUpdate));
	}
	
	//dEst = new Matrix(1, DIM, 1);
	//uEst = Matrix.identity(DIM, DIM);// initialize lower triangular part of U
	//uEst.timesEquals(6.5567);
	
	// clear chibuffer
	rejectBuf = new AccumBuffer(0, 20);
	suspectBuf = new AccumBuffer(0, 20);

	runCounter = 0; // ?
	fCov = 1; // ?

	resetted++;
    }

    public static void main(String[] args) 
    {
	// testing bierman
	int DIM = 6;
	Matrix xEst = new Matrix(1, DIM);
	Matrix uEst = Matrix.identity(DIM, DIM);// initialize lower triangular part of U
	Matrix dEst = new Matrix(1,DIM);
	Matrix PVH  = new Matrix(1, DIM);

	xEst.set(0,0, 35.06688);
	xEst.set(0,1, 12.3013);
	xEst.set(0,2, 74.11803);
	dEst.set(0,0, 0.00358);
	dEst.set(0,1, 0.00358);
	dEst.set(0,2, 0.00358);
	PVH.set(0,0, 0.46027);
	PVH.set(0,1, -0.28065);
	PVH.set(0,2, 0.84225);
	
	double R = 145.2218;
	double z = 88;
	double zin = 88;

	double c = EKFSolver.biermanProcess(z, zin, R, PVH, xEst, uEst, dEst);
	//dumpMat("xEst", xEst);
	//System.out.println("c="+c);

	//EKFSolver.copyDiag( PVH, uEst );
	//dumpMat( "PVH", PVH );

    }
}
